
#define REPEAT_RATE 13
#define REPEAT_DELAY 200

#include "usb9604.h"
#include "reg3664.h"
#include "queue16.h"
#include "ps2if.h"
#include "waitloop.h"

typedef struct Queue16 Queue;
#define qinit(Q) q16init(Q)
#define qpeek(Q) q16peek(Q)
#define qget(Q) q16get(Q)
#define qput(Q,C) q16put(Q,C)
#define qunget(Q,C) q16unget(Q,C)


#define DEBUG 1
#define DUMP 0

#define NULL 0


#define nop() asm("nop")
#define cli() asm("orc #0x80,ccr")
#define sti() asm("andc #0x7f,ccr")


#define irq1_raising_edge() (IEGR1 |= 0x02)
#define irq1_sinking_edge() (IEGR1 &= 0xfd)

#define irq2_raising_edge() (IEGR1 |= 0x08)
#define irq2_sinking_edge() (IEGR1 &= 0xf7)

#define clear_irq0() (IRR1 &= 0xfe)
#define clear_irq1() (IRR1 &= 0xfd)
#define clear_irq2() (IRR1 &= 0xfb)
#define clear_timerv_ca() (TCSRV &= 0xbf)
#define clear_timerv_cb() (TCSRV &= 0x7f)
#define clear_timerv_ov() (TCSRV &= 0xdf)

#define ps2_enable() (IENR1 |= 0x02)
#define ps2_disable() (IENR1 &= 0xfd)
//#define pc_enable() (IENR1 |= 0x04)
//#define pc_disable() (IENR1 &= 0xfb)

#define timerv_ca_enable() (TCRV0 |= 0x40)
#define timerv_ca_disable() (TCRV0 &= 0xbf)
#define timerv_ov_enable() (TCRV0 |= 0x20)
#define timerv_ov_disable() (TCRV0 &= 0xdf)

#define timer_enable() timerv_ca_enable()
#define timer_disable() timerv_ca_disable()


//Queue event_queue_l;
//Queue event_queue_h;
Queue ps2_input_queue;
Queue ps2_output_queue;

unsigned short ps2_input_bits;
unsigned short ps2_output_bits;
//unsigned short pc_input_bits;
//unsigned short pc_output_bits;

unsigned short repeat_timeout;

unsigned short ps2_byte_timeout;
unsigned short ps2_stream_timeout;



int is_accelelate_enable()
{
	return !(PDR1 & 1);
}






enum {
	KB_RESET		= 0x0000,
	KB_INIT			= 0x0100,
	KB_NORMAL		= 0x0200,
};



unsigned short mouse_state;
unsigned short mouse_init_count;

unsigned char mousetype;

int delta_x;
int delta_y;
int delta_wheel;
int buttons;

int wheel_movement;
int side_button_movement;

unsigned char mousedata[3];



#if DEBUG || DUMP

void debug_init()
{
	int i;
	PMR1 |= 0x02;
	SMR = 0;
	BRR = 25;
	SCR3 = 0;
	for(i = 0;i < 200;i++)
		asm("nop");
	SCR3 = 0x30;
	SSR = 0x80;
}


void debug_putchar(int c)
{
	while (!(SSR & 0x80));
	TDR = c;
}

void debug_print(const char *str)
{
	while (*str)
		debug_putchar(*str++);
}

void debug_putdec(int c)
{
	int n;
	char tmp[10];

	for (n = 0; c && n < 9; n++) {
		tmp[n] = '0' + c % 10;
		c /= 10;
	}
	if (!n) tmp[n++] = 0;

	while (n > 0) {
		debug_putchar(tmp[--n]);
	}
}
void debug_puthex(int c)
{
	static char hex[] = "0123456789abcdef";

	debug_putchar(hex[(c >> 4) & 15]);
	debug_putchar(hex[c & 15]);
}
void debug_putcrlf()
{
	debug_putchar('\r');
	debug_putchar('\n');
}

#endif


int countbits(unsigned short c)
{
	int i;
	for (i = 0; c; c >>= 1) {
		if (c & 1) i++;
	}
	return i;
}




//

void initialize_ps2_output()
{
	qinit(&ps2_output_queue);
	ps2_output_bits = 0;
	clear_irq1();
}

void initialize_ps2_input()
{
	qinit(&ps2_input_queue);
	ps2_input_bits = 0;
	clear_irq1();
}





void ps2_next_output(int c)
{
	int i;

	c &= 0xff;
	if ((countbits(c) & 1) == 0) c |= 0x100;	// make odd parity
	c = (c | 0x600) << 1;
	//
	// c = 000011pdddddddd0
	//                    
	//                    ^ start bit
	//            ^^^^^^^^  data
	//           ^          odd parity
	//          ^           stop bit
	//         ^            reply from mouse (ack bit)
	//
	ps2_output_bits = c;

	ps2_set_clock_0();	// I/O inhibit
	wait_40us();
	ps2_set_clock_1();	// kick interrupt

	ps2_byte_timeout = 20;
}


unsigned char ps2_last_byte;


inline void ps2_put(unsigned char c)
{
	ps2_last_byte = c;
	qput(&ps2_output_queue, c & 0xff);
#if 0
	debug_putchar('<');
	debug_puthex(c);
	debug_putcrlf();
#endif
}

inline int ps2_get()
{
	int c;
	cli();
	c = qget(&ps2_input_queue);
	sti();
#if 0
	if (c >= 0) {
		debug_putchar('>');
		debug_puthex(c);
		debug_putcrlf();
	}
#endif
	return c;
}



#pragma interrupt
void int_irq1()
{
	int c;

	if (!ps2_input_bits) {
		if (ps2_output_bits) {			// transmit mode
			if (ps2_output_bits == 1) {
				ps2_output_bits = 0;		// end transmit
			} else {
				if (ps2_output_bits & 1) {
					ps2_set_data_1();
				} else {
					ps2_set_data_0();
				}
				ps2_output_bits >>= 1;
			}
		} else {
			if (!ps2_get_data()) {
				ps2_input_bits = 0x400;		// start receive
				ps2_byte_timeout = 20;
			}
		}
	} else if (ps2_input_bits) {
		ps2_input_bits >>= 1;
		if (ps2_get_data()) {
			ps2_input_bits |= 0x400;
		}
		if (ps2_input_bits & 1) {
			if (ps2_input_bits & 0x400) {				// stop bit ?
				if (countbits(ps2_input_bits & 0x3fe) & 1) {	// odd parity ?
					c = (ps2_input_bits >> 1) & 0xff;
					qput(&ps2_input_queue, c);
				}
			}
			ps2_input_bits = 0;
		}
	}
	clear_irq1();
}


unsigned short count10 = 0;


#pragma interrupt
void int_timerv()	// 100us interval
{
	count10++;
	if (count10 >= 10) {	// every 1ms
		count10 = 0;
		if (mouse_init_count > 1) {
			mouse_init_count--;
		}
	}

	if (ps2_byte_timeout > 0) {
		ps2_byte_timeout--;
		if (ps2_byte_timeout == 0) {
			ps2_input_bits = 0;
			ps2_output_bits = 0;
			ps2_set_clock_1();
			ps2_set_data_1();
		}
	}

	if (ps2_stream_timeout > 0) {
		ps2_stream_timeout--;
	}

	//
	TCSRV &= 0x1f;
}




void handler()
{
	int c;
	long event;

	if (mouse_init_count == 1) {
		mouse_init_count = 0;
		ps2_put(0xff);
		mouse_state = KB_INIT;
		mouse_init_count = 500;
debug_putchar('!');
	}

	c = ps2_get();
	if ((mouse_state & 0xff00) == KB_INIT && c == 0xfe) {
		ps2_put(ps2_last_byte);
	} else if (c >= 0) {
		unsigned short t;
		switch (mouse_state) {
		case KB_INIT + 0:
			if (c == 0xfa) {
//debug_putchar('.');
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 1:
			if (c == 0xaa) {
//debug_putchar('.');
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 2:
//debug_putchar('.');
			// ignore c
			ps2_put(0xf2);
			mouse_state++;
			break;
		case KB_INIT + 3:
			if (c == 0xfa) {
//debug_putchar('.');
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 4:
//debug_putchar('.');
			mousetype = c;
			ps2_put(0xf3);
			mouse_state++;
			break;
		case KB_INIT + 5:
			if (c == 0xfa) {
//debug_putchar('.');
				ps2_put(200);	// sample per sec
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 6:
			if (c == 0xfa) {
//debug_putchar('.');
				ps2_put(0xe8);	// set resolution
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 7:
			if (c == 0xfa) {
//debug_putchar('.');
				ps2_put(3);	// 8count/mm
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 8:
			if (c == 0xfa) {
//debug_putchar('.');
				ps2_put(0xf4);	// enable report
				mouse_state++;
			} else {
				goto reset;
			}
			break;
		case KB_INIT + 9:
//debug_puthex(c);
//debug_putcrlf();
			if (c == 0xfa) {
				mouse_init_count = 0;
				mouse_state = KB_NORMAL;
			} else {
				goto reset;
			}
			break;
		case KB_NORMAL + 0:
			ps2_stream_timeout = 100;
			if (!(c & 0x08)) {
				break;
			}
			mousedata[0] = c;
			mouse_state++;
			break;
		case KB_NORMAL + 1:
#define MOUSE_PACKET_TIMEOUT 50		// 5ms
			if (ps2_stream_timeout == 0) {
				mousedata[0] = c;
				mouse_state = KB_NORMAL + 1;
				ps2_stream_timeout = MOUSE_PACKET_TIMEOUT;
				break;
			}
			ps2_stream_timeout = MOUSE_PACKET_TIMEOUT;
			mousedata[1] = c;
			mouse_state++;
			break;
		case KB_NORMAL + 2:
			if (ps2_stream_timeout == 0) {
				mousedata[0] = c;
				mouse_state = KB_NORMAL + 1;
				ps2_stream_timeout = MOUSE_PACKET_TIMEOUT;
				break;
			}
			mousedata[2] = c;
			mouse_state = KB_NORMAL;
			if (mousedata[0] != 0xfa) {
				// movement
				int dx, dy;
				dx = dy = 0;
				if (mousedata[0] & 0x10) {
					if (mousedata[0] & 0x40) {
						dx = (int)0xfffffeff;
					} else {
						dx = (int)0xffffff00 | mousedata[1];
					}
				} else {
					if (mousedata[0] & 0x40) {
						dx = (int)0x00000100;
					} else {
						dx = (int)(unsigned char)mousedata[1];
					}
				}
				if (mousedata[0] & 0x20) {
					if (mousedata[0] & 0x80) {
						dy = (int)0xfffffeff;
					} else {
						dy = (int)0xffffff00 | mousedata[2];
					}
				} else {
					if (mousedata[0] & 0x80) {
						dy = (int)0x00000100;
					} else {
						dy = (int)(unsigned char)mousedata[2];
					}
				}
				if (mousedata[0] & 0x04) { 
					if (abs(dx) > abs(dy)) {
						wheel_movement = 0;
						if (dx < 0) {
							buttons &= ~0x10;
							if (side_button_movement > 0) {
								side_button_movement = dx;
								buttons &= ~0x08;
							} else {
								side_button_movement += dx;
								if (side_button_movement < -100) {
									side_button_movement = -100;
									buttons |= 0x08;
								}
							}
						} else if (dx > 0) {
							buttons &= ~0x08;
							if (side_button_movement < 0) {
								side_button_movement = dx;
								buttons &= ~0x01;
							} else {
								side_button_movement += dx;
								if (side_button_movement > 100) {
									side_button_movement = 100;
									buttons |= 0x10;
								}
							}
						} else {
							buttons &= ~0x18; // clear 4th and 5th button
							side_button_movement = 0;
						}
					} else {
						buttons &= ~0x18; // clear 4th and 5th button
						side_button_movement = 0;
						wheel_movement -= dy;
						// wheel
						while (wheel_movement > 10) {
							wheel_movement -= 10;
							delta_wheel--;
						}
						while (wheel_movement < -10) {
							wheel_movement += 10;
							delta_wheel++;
						}
					}
				} else {
					buttons &= ~0x18; // clear 4th and 5th button
					side_button_movement = 0;
					wheel_movement = 0;
					if (is_accelelate_enable()) {
						int xx = dx * dx;
						int yy = dy * dy;
						if (xx + yy >= 9) {
							if (xx + yy < 300) {
								if (xx > yy) {
									if (dx < 0) {
										dx++;
									} else {
										dx--;
									}
								}
								if (xx < yy) {
									if (dy < -0) {
										dy++;
									} else {
										dy--;
									}
								}
							}
							dx *= 2;
							dy *= 2;
						}
					}
					delta_x += dx;
					delta_y -= dy;
				}
				// buttons
				buttons = (buttons & ~3) | (mousedata[0] & 0x03);
				if ((mousedata[0] & 0x06) == 0x06) { 
					buttons &= ~2;
					buttons |= 4;
				} else {
					buttons &= ~4;
				}


#if 0
debug_puthex(mousedata[0]);
debug_putchar(' ');
debug_puthex(mousedata[1]);
debug_putchar(' ');
debug_puthex(mousedata[2]);
debug_putcrlf();
#endif
			}
			break;
		}
		
	}

	return;

reset:;
	mouse_state = KB_RESET;
	mouse_init_count = 10;
	debug_putchar('?');
	debug_puthex(c);
	debug_putcrlf();
}


static inline void usbn_write_address(int a)
{
	PDR8 = a;
	PCR8 = 0xff;
	PDR5 &= ~1;		// ~CS = 0
	PDR5 |= 8;		// A0 = 1
	PDR5 &= ~4;		// ~WR = 0
	nop();
	PDR5 |= 4;		// ~WR = 1
	nop();
	PCR8 = 0x00;
	PDR5 |= 1;		// ~CS = 1
}

static inline int usbn_read_data()
{
	int d;
	PCR8 = 0x00;
	PDR5 &= ~1;		// ~CS = 0
	PDR5 &= ~8;		// A0 = 0
	PDR5 &= ~2;		// ~RD = 0
	nop();
	d = PDR8;
	PDR5 |= 2;		// ~RD = 1
	PDR5 |= 1;		// ~CS = 1
	return d;
}

static inline void usbn_write_data(int d)
{
	PDR8 = d;
	PCR8 = 0xff;
	PDR5 &= ~1;		// ~CS = 0
	PDR5 &= ~8;		// A0 = 0
	PDR5 &= ~4;		// ~WR = 0
	nop();
	PDR5 |= 4;		// ~WR = 1
	nop();
	PDR5 |= 1;		// ~CS = 1
	PCR8 = 0x00;
}

static inline void usbn_clear_stall(int epc)
{
	usbn_write(epc, usbn_read(epc) & 0x7f); // clear STALL
}

static inline void usbn_set_stall(int epc)
{
	usbn_write(epc, usbn_read(epc) | 0x80); // set STALL
}

int usbn_read(int a)
{
	usbn_write_address(a);
	return usbn_read_data();
}

int usbn_write(int a, int d)
{
	usbn_write_address(a);
	usbn_write_data(d);
}


void usbn_flush_tx0()
{
	usbn_write(USBN_TXC0, 0x80);
}

void usbn_enable_rx0()
{
	usbn_write(USBN_RXC0, usbn_read(USBN_RXC0) | 0x01);
}

static const unsigned char device_desc[] = {
	18,
	1,
	0, // lo: 0x0100
	1, // hi: 0x0100
	0,
	0,
	0,
	8,
	0x5e, // microsoft
	0x04,
	0x09, // intellimouse
	0x00,
	0,
	0,
	0,
	0,
	0,
	1,
};

struct configure_t {

	const unsigned char config_desc[9];
	const unsigned char iface_desc[9];
	const unsigned char ep1_desc[7];
	const unsigned char hid_desc[9];

} configure = {
	// config_desc
		9,
		2,
		9 + 9 + 9 + 7,
		0,
		1,
		1,
		4,
		0x80,
		0x32, // 100mA
	// char iface_desc
		9,
		4,
		0,
		0,
		1,
		3,
		1,
		2,
		0,
	// hid_desc
		9,
		0x21,
		0x00, // lo: 0x0100
		0x01, // hi: 0x0100
		0,
		1,
		0x22,
		46,   // lo: sizeof(rep_desc)
		0,    // hi:
	// ep1_desc
		7,
		5,
		0x81,
		3,
		4,    // packet size
		0,
		10,
};


static const unsigned char rep_desc[] = {
	0x05, 0x01, // Usage Oage (Generic Desktop)
	0x09, 0x02, // Usage (Mouse)
	0xa1, 0x01, // Collection (Application)
	0x09, 0x01, //   Usage (Pointer)
	0xa1, 0x00, //   Collection (Physical)
	0x05, 0x09, //     Usage Page (Buttons)
	0x19, 0x01, //     Usage Minimum (01)
	0x29, 0x08, //     Usage Maximum (08)
	0x15, 0x00, //     Logical Minimum (0)
	0x25, 0x01, //     Logical Maximum (1)
	0x95, 0x08, //     Report Count (8)
	0x75, 0x01, //     Report Size (1)
	0x81, 0x02, //     Input (Data, Variable, Absolute)
	0x05, 0x01, //     Usage Page (Generic Desktop)
	0x09, 0x30, //     Usage (X)
	0x09, 0x31, //     Usage (Y)
	0x09, 0x38, //     Usage (Wheel)
	0x15, 0x81, //     Logical Minimum (-127)
	0x25, 0x7f, //     Logical Maximum (127)
	0x75, 0x08, //     Report Size (8)
	0x95, 0x03, //     Report Size (3)
	0x81, 0x06, //     Input (Data, Variable, Relative)
	0xc0,       //   End Collection
	0xc0,       // End Collection
};

static const unsigned char str_desc[] = {
	4,
	3,
	9,
	0,

	10,
	3,
	'H', 0,
	'o', 0,
	'g', 0,
	'e', 0,
};

unsigned char const *tx0ptr = 0;
unsigned int tx0len = 0;

unsigned char usbconfig = 0;
unsigned char usb_ep0_data01 = 0;
unsigned char usb_ep1_data01 = 1;

void senddesc()
{
	if (tx0len > 0) {
		unsigned int i, n;
		n = tx0len;
		if (n > 8) {
			n = 8;
		}
		for (i = 0; i < n; i++) {
			usbn_write(USBN_TXD0, tx0ptr[i]);
		}
		tx0ptr += n;
		tx0len -= n;
	}
}

void usbn_reset()
{
	usbn_write(USBN_MCNTRL, 0x05);
	wait_100ms();
	usbn_write(USBN_MCNTRL, 0xc4);
	usbn_write(USBN_CCONF, 4);
}

void usbn_setup()
{
	usbn_write(USBN_FAR, 0x80);
	usbn_write(USBN_EPC0, 0x00);
	usbn_write(USBN_EPC1, 0x11);
	usbn_write(USBN_NAKMSK, 0x10);
	usbn_write(USBN_TXMSK, 0x03);
	usbn_write(USBN_RXMSK, 0x01);
	usbn_write(USBN_ALTMSK, 0x40);
	usbn_write(USBN_MAMSK, 0x56); // INTR | RX_EV | NAK | TX_EV | ALT
	usbn_write(USBN_TXC0, 0x08); // FLUSH
	usbn_write(USBN_RXC0, 0x01); // RX_EN
	tx0ptr = 0;
	tx0len = 0;
}

void usbn_wakeup()
{
	usbn_write(USBN_NFSR, 0x02); // operational
	usbn_write(USBN_MCNTRL, 0xcc);
}


void usbn_run()
{
	unsigned char maev = usbn_read(USBN_MAEV);

	if (maev & 0x10) { // NAK
		unsigned char nakev = usbn_read(USBN_NAKEV);
		if (nakev & 0x01) { // in 0
			if (tx0len > 0) {
				usbn_write(USBN_TXC0, 0x08); // flush
				usbn_write(USBN_RXC0, 0x01); // rx_en
				tx0ptr = 0;
				tx0len = 0;
			}
		}
		if (nakev & 0x02) { // in 1
		}
		if (nakev & 0x04) { // in 2
		}
		if (nakev & 0x08) { // in 4
		}
		if (nakev & 0x10) { // out 0
			if (tx0len > 0) {
				usbn_write(USBN_TXC0, 0x08); // flush
				usbn_write(USBN_RXC0, 0x01); // rx_en
				tx0ptr = 0;
				tx0len = 0;
			}
		}
		if (nakev & 0x20) { // out 1
		}
		if (nakev & 0x40) { // out 2
		}
		if (nakev & 0x80) { // out 3
		}
	}
	if (maev & 0x40) { // RX_EV
		unsigned char rxev = usbn_read(USBN_RXEV);
		if (rxev & 0x01) { // EP0
			unsigned char rxs0 = usbn_read(USBN_RXS0);
			if (rxs0 & 0x40) { // setup
				int i;
				unsigned char req[8];
				for (i = 0; i < 8; i++) {
					req[i] = usbn_read(USBN_RXD0);
				}
				usbn_write(USBN_RXC0, 0x08);
				usbn_write(USBN_TXC0, 0x08);
				usbn_clear_stall(USBN_EPC0);
				switch (req[0] & 0x60) { // type
				case 0x00: // standard
					switch (req[1]) {
					case 0: // get status
						usbn_write(USBN_TXD0, 0);
						usbn_write(USBN_TXD0, 0);
						break;
					case 1: // clear feature
						switch (req[2] & 3) {
						case 2:
							switch (req[4] & 0x0f) {
							case 0: usbn_clear_stall(USBN_EPC0); break;
							case 1: usbn_clear_stall(USBN_EPC1); break;
							case 2: usbn_clear_stall(USBN_EPC2); break;
							case 3: usbn_clear_stall(USBN_EPC3); break;
							case 4: usbn_clear_stall(USBN_EPC4); break;
							case 5: usbn_clear_stall(USBN_EPC5); break;
							case 6: usbn_clear_stall(USBN_EPC6); break;
							}
						}
						break;
					case 3: // set feature
						break;
					case 5: // set address
						usbn_write(USBN_EPC0, 0x40);
						usbn_write(USBN_FAR, req[2] | 0x80);
						break;
					case 6: // get descriptor
						switch (req[3]) {
						case 1: // device
							tx0ptr = device_desc;
							tx0len = sizeof(device_desc);
							break;
						case 2: // configuration
							tx0ptr = (unsigned char const *)&configure;
							tx0len = sizeof(configure);
							break;
						case 3: // string
							tx0ptr = str_desc;
							tx0len = sizeof(str_desc);
							break;
						case 4: // interface
							tx0ptr = configure.iface_desc;
							tx0len = sizeof(configure.iface_desc);
							break;
//						case 5: // endpoint
//							tx0ptr = configure.ep0_desc;
//							tx0len = sizeof(configure.ep0_desc);
//							break;
						case 0x21: // HID
							tx0ptr = configure.hid_desc;
							tx0len = sizeof(configure.hid_desc);
							break;
						case 0x22: // HID report
							tx0ptr = rep_desc;
							tx0len = sizeof(rep_desc);
							break;
						default:
							tx0ptr = 0;
							tx0len = 0;
							break;
						}
						if (tx0ptr && tx0len > 0) {
							int n = req[6] | (req[7] << 8);
							if (tx0len > n) {
								tx0len = n;
							}
						}
						senddesc();
						break;
					case 8: // get configuration
            			usbn_write(USBN_TXD0, usbconfig);
						break;
					case 9: // set configuration
						break;
					case 10: // get interface
	                    usbn_write(USBN_TXD0, 0);
						break;
					case 11: // set interface
						if (req[2] != 0) {
							usbn_set_stall(USBN_EPC0);
						}
						break;
					default:
						usbn_set_stall(USBN_EPC0);
						break;
					}
					break;
				case 0x20: // class
	                switch (req[1]) {
                    case 1: // get report
//								getreport();
						break;
                    case 2: // get idle
//								getidle();
						break;
                    case 3: // get protocol
//								getprotocol();
						break;
                    case 9: // set report
//								setreport();
						break;
                    case 10: // set idle
//								setidle();
						break;
                    case 11: // set protocol
//								setprotocol();
						break;
                    default:
						usbn_set_stall(USBN_EPC0);
						break;
	                }
					break;
//						case 0x40: // vendor
//							break;
//						case 0x60: // reserved
//							break;
				default:
					usbn_set_stall(USBN_EPC0);
					break;
				}
				usbn_write(USBN_TXC0, 0x05); // toggle + tx_en
				usb_ep0_data01 = 0;
			} else { // not setup
				if (tx0len > 0) {
					usbn_write(USBN_TXC0, 0x08); // flush
					tx0ptr = 0;
					tx0len = 0;
				} else {
					usbn_write(USBN_RXC0, 0x08); // flush
					usbn_write(USBN_RXC0, 0x01); // rx_en
				}
			}
		}
		if (rxev & 0x02) { // EP2
		    unsigned char rxs3 = usbn_read(USBN_RXS1);
		    usbn_write(USBN_RXC1, 0x08);
		    usbn_write(USBN_RXC1, 0x01);
		}
		if (rxev & 0x04) { // EP4
		    unsigned char rxs3 = usbn_read(USBN_RXS2);
		    usbn_write(USBN_RXC2, 0x08);
		    usbn_write(USBN_RXC2, 0x01);
		}
		if (rxev & 0x08) { // EP6
		    unsigned char rxs3 = usbn_read(USBN_RXS3);
		    usbn_write(USBN_RXC3, 0x08);
		    usbn_write(USBN_RXC3, 0x01);
		}
	}
	if (maev & 0x04) { // TX_EV
		unsigned char txev = usbn_read(USBN_TXEV);
		if (txev & 0x01) { // EP0
			unsigned char txs0 = usbn_read(USBN_TXS0);
			if ((txs0 & 0x60) == 0x60) {
				usbn_write(USBN_TXC0, 0x08); // flush
				if (tx0len > 0) {
					senddesc();
					if (usb_ep0_data01 == 0) {
						usbn_write(USBN_TXC0, 0x01); // tx_en
						usb_ep0_data01 = 1;
					} else {
						usbn_write(USBN_TXC0, 0x05); // toggle + tx_en
						usb_ep0_data01 = 0;
					}
				} else {
					usbn_write(USBN_RXC0, 0x01);
				}
			} else {
				tx0ptr = 0;
				tx0len = 0;
				usbn_write(USBN_RXC0, 0x01);
			}
		}
		if (txev & 0x02) { // EP1
			unsigned char txs1 = usbn_read(USBN_TXS1);
			if ((txs1 & 0x60) == 0x60) {
				int dx, dy;
				if (delta_x < -127) {
					delta_x += 127;
					dx = -127;
				} else if (delta_x > 127) {
					delta_x -= 127;
					dx = 127;
				} else {
					dx = delta_x;
					delta_x = 0;
				}
				if (delta_y < -127) {
					delta_y += 127;
					dy = -127;
				} else if (delta_y > 127) {
					delta_y -= 127;
					dy = 127;
				} else {
					dy = delta_y;
					delta_y = 0;
				}
				usbn_write(USBN_TXC1, 0x08); // flush
				usbn_write(USBN_TXD1, buttons);
				usbn_write(USBN_TXD1, dx);
				usbn_write(USBN_TXD1, dy);
				usbn_write(USBN_TXD1, delta_wheel);
				delta_wheel = 0;
				if (usb_ep1_data01 == 0) {
					usbn_write(USBN_TXC1, 0x07); // toggle + last + tx_en
					usb_ep1_data01 = 1;
				} else {
					usbn_write(USBN_TXC1, 0x03); // last + tx_en
					usb_ep1_data01 = 0;
				}
			}
		}
		if (txev & 0x04) { // EP3
			unsigned char txs2 = usbn_read(USBN_TXS2);
		}
		if (txev & 0x08) { // EP5
			unsigned char txs3 = usbn_read(USBN_TXS3);
		} 
	}
	if (maev & 0x02) { // ALT
		unsigned char altev = usbn_read(USBN_ALTEV);
		if (altev & 0x40) { // reset
			usbn_setup();
			usbn_wakeup();
			tx0ptr = 0;
			tx0len = 0;
			while (usbn_read(USBN_ALTEV) & 0x40) {
			}
		} else if (altev & 0x10) { // SD3
			usbn_write(USBN_ALTMSK, 0xc0);
			usbn_write(USBN_NFSR, 3);
		} else if (altev & 0x80) { // RESUME
			usbn_write(USBN_ALTMSK, 0x50);
			usbn_write(USBN_NFSR, 2);
		}
	}
}

int main()
{
	int c;
	unsigned short t;

	// external interrupt pins

	PMR1 = 0x30;
	PCR1 = 0x00;
	PUCR1 = 0x0f;

	// I/O port pins

	PDR8 = 0x00;
	PCR8 = 0x00;

	PMR5 = 0x00;
	PDR5 = 0x0f;
	PCR5 = 0x5f;

	// setup external interrupt

	IEGR1 = 0x00;
	IENR1 = 0x10;

	// setup 10kHz interval timer

	TIOR1 = 0x00;
	TCRV0 = 0x0a;
	TCRV1 = 0xe2;
	TCORA = 99;			// 16000000 / 16 / 100 = 10000

#if DEBUG || DUMP
	// for debugging ports

#endif

	//

	ps2_set_clock_0();
	ps2_set_data_1();

//	qinit(&event_queue_l);
//	qinit(&event_queue_h);
	initialize_ps2_input();
	initialize_ps2_output();

	mousetype = 0;

	delta_x = 0;
	delta_y = 0;
	buttons = 0;
	wheel_movement = 0;
	side_button_movement = 0;

	mouse_init_count = 200;		// delay for initialize
	mouse_state = KB_RESET;
	ps2_byte_timeout = 0;
	ps2_stream_timeout = 0;

	ps2_set_clock_1();
	timer_enable();
	ps2_enable();

#ifdef DEBUG
	debug_init();
	debug_print("\r\nPS2USB\r\n");
#endif

	usbn_reset();
	usbn_setup();
	usbn_wakeup();

	// kick interrupt transfer
	usbn_read(USBN_TXS1);
	usbn_write(USBN_TXC1, 0x08);
	usbn_write(USBN_TXD1, 0);
	usbn_write(USBN_TXD1, 0);
	usbn_write(USBN_TXD1, 0);
	usbn_write(USBN_TXD1, 0);
	usbn_write(USBN_TXC1, 0x03);


	IENR1 |= 0x02;
	IRR1 = 0x00;
	sti();


#if 0


	while (1) {
		usbn_run();
	}


#else

	while (1) {
		usbn_run();

		handler();

		// transmit to mouse
		if (!ps2_input_bits) {
			int c = qget(&ps2_output_queue);
			if (c >= 0) {
				ps2_next_output(c);
			}
		}
	}
	
#endif

	return 0;
}





